#

library(data.tree)
library(networkD3)
library(dplyr)
library(DEoptim)


create_unique_ids <- function(n, seed_no = 1, char_len = 5) {
  set.seed(seed_no)
  pool <- c(letters, LETTERS, 0:9)

  res <- character(n) # pre-allocating vector is much faster than growing it
  for (i in seq(n)) {
    this_res <- paste0(sample(pool, char_len, replace = TRUE), collapse = "")
    while (this_res %in% res) { # if there was a duplicate, redo
      this_res <- paste0(sample(pool, char_len, replace = TRUE), collapse = "")
    }
    res[i] <- this_res
  }
  res
}

CellOnDividing <- function(D_G1, D_S, D_G2, D_M, a, b, k, alph, V_0, Inht, InitSize, Pday = 0, PTime = -1) {
  # concentration of inhibitor in initial cell

  C_init <- Inht / InitSize
  # growth speed for G1 step
  V_init <- a * exp(-k * C_init) + b #micron per day
  # time for every step
  V_G1_time <- (D_G1 - InitSize) / V_init
  V_S_time <- (D_S - D_G1) / V_0
  V_G2_time <- (D_G2 - D_S) / V_0
  D_M_time <- (D_M - D_G2) / V_0

  if (PTime >= 0) {
    BirthDay <- Pday + PTime
    initCell <- list(
      C = C_init, V_G1_time = V_G1_time, V_S_time = V_S_time, V_G2_time = V_G2_time, D_M_time = D_M_time,
      Inht = Inht, V_init = V_init, SeasonBirthDay = BirthDay, BirthGivingTime = PTime
    )
  }
  else {
    initCell <- list(
      C = C_init, V_G1_time = V_G1_time, V_S_time = V_S_time, V_G2_time = V_G2_time, D_M_time = D_M_time,
      Inht = Inht, V_init = V_init, SeasonBirthDay = 0, BirthGivingTime = 0
    )
  }

  return(initCell)
}

IsAnyToDivide <- function(CellDivisionAsDataFrame, C_min) {
  cells <- data.frame()
  cells <- CellDivisionAsDataFrame
  cells <- subset(cells, InhtConc >= C_min)
  if (nrow(cells) > 0) {
    return(TRUE)
  }
  else {
    return(FALSE)
  }
}




BuildRibbonStructureFromTree <- function(treeDF, DividingsEnabled = 100, GrowthSeasonInDays = 120) {
 
  treeDF_daughters <- subset(treeDF, Priority == 2 & TFromInitial <= GrowthSeasonInDays)
  treeDF_daughters <- treeDF_daughters[order(treeDF_daughters$TFromInitial), ]

  
  HistoryList <- list()

  currDividing <- list()
  source <- c()
  target <- c()
  Inht <- c()
  TCum <- c()
  V_init <- c()



  for (j in 1:nrow(treeDF_daughters))
  {
    target <- c(target, treeDF_daughters[j, ]$ID)
    Inht <- c(Inht, treeDF_daughters[j, ]$InhtAmount)
    TCum <- c(TCum, treeDF_daughters[j, ]$T_Cumulative)
    V_init <- c(V_init, treeDF_daughters[j, ]$V_init)

    
    if ((length(source) == 0) || (is.na(match(as.character(treeDF_daughters[j, ]$Parent), source)) == FALSE) ||
      (is.na(match(as.character(treeDF_daughters[j, ]$Parent), target)) == FALSE)) {
      source <- c(source, as.character(treeDF_daughters[j, ]$Parent))
    }
    else {
     
      path <- strsplit(as.character(treeDF_daughters[j, ]$pathString), "/")[[1]]

      for (p in (length(path) - 1):2)
      {
        tmp_cell <- subset(treeDF, ID == path[p])
        
        if (tmp_cell$Priority == 2) 

          {
            source <- c(source, as.character(tmp_cell$ID))
            
            break
          }
      }
    }
    
    
  }
  currDividing$source <- source

  currDividing$target <- target

 
  links <- data.frame(
    source = currDividing$source,
    target = currDividing$target,
    inht = Inht,
    tCum = TCum,
    V_init = V_init,
    value = rep(1, length(source)),
    state = rep("old", length(source))
  )
 
  HistoryList <- list()
  for (n in 1:nrow(links))
  {
    tmp <- head(links, n)
    
    if (length(unique(tmp$source)) == n) {
      state <- c(rep("old", n - 1), "new")
      tmp <- subset(tmp, select = -c(state))
      HistoryList[[n]] <- cbind(tmp, state)
    }
    else {
      old_data <- HistoryList[[n - 1]]

      toCheck <- links[n, ]
      if (is.na(match(toCheck$source[1], old_data$source)) == FALSE) #new string replaces old values
        {
          curr_row <- subset(old_data, source == as.character(toCheck$source[1]))
         
          new_row <- data.frame(
            c(as.character(curr_row$source[1])), c(as.character(toCheck$target[1])),
            c(toCheck$inht[1]), c(toCheck$tCum[1]), c(toCheck$V_init[1]), c(1), "new"
          )
         
          names(new_row) <- names(old_data)
         
          new_row1 <- data.frame(
            c(as.character(toCheck$target[1])), c(as.character(curr_row$target[1])),
            c(curr_row$inht[1]), c(curr_row$tCum[1]), c(curr_row$V_init[1]), c(1), "old"
          )
          names(new_row1) <- names(old_data)
          old_data <- subset(old_data, source != as.character(toCheck$source[1]))
          old_data <- rbind(old_data, new_row)
          old_data <- rbind(old_data, new_row1)
          state <- c(rep("old", nrow(old_data) - 2), "new", "old")
          old_data <- subset(old_data, select = -c(state))
          HistoryList[[n]] <- cbind(old_data, state)
        }
      else {
        new_row <- data.frame(
          c(as.character(toCheck$source[1])), c(as.character(toCheck$target[1])),
          c(toCheck$inht[1]), c(toCheck$tCum[1]), c(toCheck$V_init[1]), c(1), "new"
        )
        names(new_row) <- names(old_data)
        old_data <- rbind(old_data, new_row)
        state <- c(rep("old", nrow(old_data) - 1), "new")
        old_data <- subset(old_data, select = -c(state))
        HistoryList[[n]] <- cbind(old_data, state)
        # merge old and checked
      }
    }
  }

  return(HistoryList)
}



# returns all divings with no limits
CalculateCellsDivisionWithParents_VSControlled <- function(D_G1, D_S, D_G2, D_M, a, b, k, alph, V_0, Inht, InitSize, C_min, GRates, s = 1) {
  n <- s #set stress inhibitor coefficient
 

  ids <- create_unique_ids(1000)
  IDiterator <- 1
  names <- c(
    "ID", "G1_time", "S_time", "G2_time", "M_time", "InhtAmount", "InhtConc", "V_init", "Parent",
    "T_Cumulative", "Priority", "BirthDay", "BirthGivingTime"
  )
  cells <- data.frame(matrix(ncol = length(names), nrow = 0))
  fullCellData <- data.frame(matrix(ncol = length(names), nrow = 0))
  colnames(cells) <- names
  colnames(fullCellData) <- names
  C_curr <- Inht / InitSize # здесь количество ингибитора, а не его концентрация!
  initCell <- CellOnDividing(D_G1, D_S, D_G2, D_M, a, b, k, alph, V_0, Inht, InitSize, Pday = 0, PTime = 0)

  cell_curr <- data.frame(
    0, initCell$V_G1_time, initCell$V_S_time, initCell$V_G2_time, initCell$D_M_time,
    initCell$Inht, initCell$C, initCell$V_init, " ",
    initCell$V_G1_time + initCell$V_S_time + initCell$V_G2_time + initCell$D_M_time, 2,
    initCell$SeasonBirthDay, initCell$BirthGivingTime
  )
  colnames(cell_curr) <- names
  cells <- rbind(cells, cell_curr)
  IsAnyCellsOnDivide <- IsAnyToDivide(cells, C_min)
  while (IsAnyCellsOnDivide) {
    new_cells <- data.frame(matrix(ncol = length(names), nrow = 0))
    colnames(new_cells) <- names
    for (i in 1:nrow(cells))
    {
      if (cells[i, ]$InhtConc >= C_min) {
        cell_ReadyToDivide <- CellOnDividing(D_G1, D_S, D_G2, D_M, a, b, k, alph, V_0, cells[i, ]$InhtAmount, D_M / 2, PTime = cells[i, ]$T_Cumulative, Pday = cells[i, ]$BirthDay)
        

        # check for stress

        if (cell_ReadyToDivide$SeasonBirthDay > 3 & cell_ReadyToDivide$SeasonBirthDay < nrow(GRates)) {
          # if stress, then less inhibitor amount
          if ((GRates[round(cell_ReadyToDivide$SeasonBirthDay), ]$Gr - GRates[round(cell_ReadyToDivide$SeasonBirthDay) - 1, ]$Gr) < 0 &
            (GRates[round(cell_ReadyToDivide$SeasonBirthDay), ]$Gr - GRates[round(cell_ReadyToDivide$SeasonBirthDay) - 2, ]$Gr) < 0) {
            
            cell_daughter1 <- data.frame(
              ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
              cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
              cell_ReadyToDivide$Inht * alph * n, (cell_ReadyToDivide$Inht * alph * n) / (D_M / 2),
              cell_ReadyToDivide$V_init, cells[i, ]$ID,
              
              cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 1,
              cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
            )
            IDiterator <- IDiterator + 1
            cell_daughter2 <- data.frame(
              ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
              cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
              cell_ReadyToDivide$Inht * (1 - alph) * n, (cell_ReadyToDivide$Inht * (1 - alph) * n) / (D_M / 2), cell_ReadyToDivide$V_init, cells[i, ]$ID,
              cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 2,
              cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
            )
          }

          else #else - common calculation
          {
            
            cell_daughter1 <- data.frame(
              ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
              cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
              cell_ReadyToDivide$Inht * alph, (cell_ReadyToDivide$Inht * alph) / (D_M / 2), cell_ReadyToDivide$V_init, cells[i, ]$ID,
             
              cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 1,
              cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
            )
            IDiterator <- IDiterator + 1
            cell_daughter2 <- data.frame(
              ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
              cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
              cell_ReadyToDivide$Inht * (1 - alph), (cell_ReadyToDivide$Inht * (1 - alph)) / (D_M / 2), cell_ReadyToDivide$V_init, cells[i, ]$ID,
              cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 2,
              cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
            )
          }
        }

        else {


          # next step id generate 2 daughters
          
          cell_daughter1 <- data.frame(
            ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
            cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
            cell_ReadyToDivide$Inht * alph, (cell_ReadyToDivide$Inht * alph) / (D_M / 2), cell_ReadyToDivide$V_init, cells[i, ]$ID,
           
            cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 1,
            cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
          )

          IDiterator <- IDiterator + 1
          cell_daughter2 <- data.frame(
            ids[IDiterator], cell_ReadyToDivide$V_G1_time, cell_ReadyToDivide$V_S_time,
            cell_ReadyToDivide$V_G2_time, cell_ReadyToDivide$D_M_time,
            cell_ReadyToDivide$Inht * (1 - alph), (cell_ReadyToDivide$Inht * (1 - alph)) / (D_M / 2), cell_ReadyToDivide$V_init, cells[i, ]$ID,
            cell_ReadyToDivide$V_G1_time + cell_ReadyToDivide$V_S_time + cell_ReadyToDivide$V_G2_time + cell_ReadyToDivide$D_M_time, 2,
            cell_ReadyToDivide$SeasonBirthDay, cell_ReadyToDivide$BirthGivingTime
          )
        }
        IDiterator <- IDiterator + 1
        colnames(cell_daughter1) <- names
        colnames(cell_daughter2) <- names
        new_cells <- rbind(new_cells, cell_daughter1)
        new_cells <- rbind(new_cells, cell_daughter2)
      }
      else {
        new_cells <- rbind(new_cells, cells[i, ])
      }
    }
    
    IsAnyCellsOnDivide <- IsAnyToDivide(new_cells, C_min)
    
    cells <- data.frame()
    cells <- new_cells
    fullCellData <- rbind(fullCellData, cells)
  }
  fullCellData <- rbind(cell_curr, fullCellData)

 
  fullCellData <- distinct(fullCellData, ID, .keep_all = TRUE)

  PathToParent <- c()
  TFromInitial <- c()
  for (i in 1:nrow(fullCellData))
  {
    tmp_path <- paste(
      fullCellData[i, ]$Parent,
      fullCellData[i, ]$ID,
      sep = "/"
    )
    tmp_time <- 0 
    currParent <- fullCellData[i, ]$Parent
    for (j in i:1)
    {
      if (fullCellData[j, ]$ID == currParent) {
        tmp_path <- paste(fullCellData[j, ]$Parent, tmp_path, sep = "/")
        tmp_time <- tmp_time + fullCellData[j, ]$T_Cumulative
        currParent <- fullCellData[j, ]$Parent
      }
    }
    tmp_path <- paste(tmp_path, sep = "/")
    PathToParent <- c(PathToParent, tmp_path)
    TFromInitial <- c(TFromInitial, tmp_time)
    tmp_path <- ""
    tmp_time <- 0

  }
  fullCellData <- cbind(fullCellData, PathToParent)
  fullCellData <- cbind(fullCellData, TFromInitial)
  names(fullCellData) <- c(names, c("pathString", "TFromInitial"))


  

  return(fullCellData)
}




HowManyTerminalsAchieved <- function(treeDF, GrowthSeasonInDays = 120,
                                     GrowthSeasonStartDay = 150, C_min = 0.018) {
  treeDF <- treeDF[order(treeDF$TFromInitial), ]
  terminals <- subset(treeDF, TFromInitial <= GrowthSeasonInDays & InhtConc <= C_min)

  return(terminals)
}

# builds ribbon as data frame to make curves and analyse data
# returns RibbonVector (see next function)
BuildRibbonStructureForAnalasys <- function(HistoryList, treeData, customIndex = 0) {
  Ribbon <- list()
  Ribbon <- HistoryList # input data is a list of dividings, presented as data frames, and dividing tree data as DF
  cellDivision <- c()
  cellDivisionActualChanges <- c()
  StateVector <- c()

  myIndex <- customIndex
  if (customIndex == 0 || customIndex > length(Ribbon)) {
    myIndex <- length(Ribbon)
  }
  NewItem <- subset(Ribbon[[myIndex]], state == "new")$source
  print(NewItem)
  print(Ribbon)
  for (i in 1:myIndex)
  {
    new_cell <- subset(Ribbon[[i]], state == "new")
    parent <- as.character(new_cell$source)
    child <- as.character(new_cell$target)
    if (length(cellDivision) == 0) {
      cellDivision <- c(cellDivision, child)
      cellDivisionActualChanges <- c(cellDivisionActualChanges, child)
    }

    else {
     

      position <- match(c(parent), cellDivision)
      



      cellDivision <- c(cellDivision[1:position], child, cellDivision[(position + 1):length(cellDivision)])
      cellDivision <- cellDivision[!is.na(cellDivision)]
     
      cellDivision <- unique(cellDivision)

      # find information for mother cell

      # new_parent_info
      child_info <- subset(treeData, ID == child)
      path <- strsplit(as.character(child_info$pathString), "/")[[1]]
      parent_info <- subset(treeData, Parent == path[length(path) - 1])
      parent_info <- subset(parent_info, Priority == 1)

      if (position == 1) {
        cellDivisionActualChanges <- c(parent_info$ID, child, cellDivisionActualChanges[2:length(cellDivisionActualChanges)])
      }

      else if (position == 2) {
        cellDivisionActualChanges <- c(cellDivisionActualChanges[1], parent_info$ID, child, cellDivisionActualChanges[3:length(cellDivisionActualChanges)])
      }
      else if (position == length(cellDivision) & position > 2) {
        cellDivisionActualChanges <- c(cellDivisionActualChanges[1:(position - 1)], parent_info$ID, child)
      }
      else {
        cellDivisionActualChanges <- c(cellDivisionActualChanges[1:(position - 1)], parent_info$ID, child, cellDivisionActualChanges[(position + 1):length(cellDivisionActualChanges)])
      }




     
    }
  }

  cellDivisionActualChanges <- cellDivisionActualChanges[!is.na(cellDivisionActualChanges)] # если значения в предыдущей строке выходят за границы массива, возникают начени NA и массив повторяется сначала. Я не разобралась почему
  cellDivisionActualChanges <- cellDivisionActualChanges[1:length(cellDivision)]

  tmp_list <- list()
  tmp_list$cellDivision <- cellDivision
  tmp_list$cellDivisionActualChanges <- cellDivisionActualChanges

  NewItemPos <- match(as.character(NewItem), cellDivision) # now we know position of "new" state
  NewItemPos <- NewItemPos + 1 
 
  StateVector <- c()

  
  StateVector <- c(rep("old", NewItemPos - 1), "new", rep("old", length(cellDivisionActualChanges) - NewItemPos))
 


  tmp_list$state <- StateVector


  return(tmp_list)
}


GetDataForRibbonStructure <- function(RibbonVector, treeData) {
  df <- treeData[FALSE, ] # empty df
  for (i in 1:length(RibbonVector))
  {
    tmp <- subset(treeData, ID == RibbonVector[i])
    df <- rbind(df, tmp)
  }
  return(df)
}
